from EEETools.MainModules import Block
from EEETools.MainModules.support_blocks import Drawer
import xml.etree.ElementTree as ETree


class Alternator(Block):

    def __init__(self, inputID, main_class):

        super().__init__(inputID, main_class)

        self.type = "alternator"
        self.has_support_block = True
        self.support_block.append(Drawer(main_class, self))

        self.efficiency = 1.

    def add_connection_to_support_block(self, new_connection, is_input):
        self.support_block[0].add_connection(new_connection, is_input, )

    def is_ready_for_calculation(self):
        return len(self.support_block[0].input_connections) >= 1

    def prepare_for_calculation(self):

        self.support_block[0].prepare_for_calculation()

        new_conn = self.main_class.append_connection(from_block=self)
        new_conn.name = "electrical power output"
        new_conn.is_useful_effect = True
        new_conn.automatically_generated_connection = True
        new_conn.exergy_value = self.exergy_balance

    def append_excel_connection_list(self, input_list):

        self.efficiency = float(input_list[0])

        for elem in input_list[1:]:

            new_conn = self.main_class.find_connection_by_index(abs(elem))

            if not new_conn is None:

                is_input = (elem > 0)
                new_conn.is_fluid_stream = False

                if is_input:

                    self.add_connection(new_conn, is_input, append_to_support_block=0)

                else:

                    self.add_connection(new_conn, is_input)

    def export_xml_other_parameters(self) -> ETree.Element:

        other_tree = ETree.Element("Other")
        other_tree.set("efficiency", str(self.efficiency))
        return other_tree

    def append_xml_other_parameters(self, input_list: ETree.Element):

        self.efficiency = float(input_list.get("efficiency"))

    def export_xml_connection_list(self) -> ETree.Element:

        xml_connection_list = ETree.Element("Connections")

        mechanical_connections = ETree.SubElement(xml_connection_list, "MechanicalConnections")

        for input_connection in self.support_block[0].external_input_connections:

            if not input_connection.automatically_generated_connection:

                input_xml = ETree.SubElement(mechanical_connections, "input")
                input_xml.set("index", str(input_connection.index))

        for output_connection in self.support_block[0].external_output_connections:

            if not output_connection.automatically_generated_connection:

                output_xml = ETree.SubElement(mechanical_connections, "output")
                output_xml.set("index", str(output_connection.index))

        electrical_connections = ETree.SubElement(xml_connection_list, "ElectricalConnections")

        for input_connection in self.external_input_connections:

            if not input_connection.automatically_generated_connection:

                input_xml = ETree.SubElement(electrical_connections, "input")
                input_xml.set("index", str(input_connection.index))

        for output_connection in self.external_output_connections:

            if not output_connection.automatically_generated_connection:

                output_xml = ETree.SubElement(electrical_connections, "output")
                output_xml.set("index", str(output_connection.index))

        a = ETree.tostring(xml_connection_list)
        return xml_connection_list

    def append_xml_connection_list(self, input_list: ETree.Element):

        a = ETree.tostring(input_list)

        mechanical_connections = input_list.find("MechanicalConnections")
        electrical_connections = input_list.find("ElectricalConnections")

        self.__add_connection_by_index(mechanical_connections, "input", append_to_support_block=0)
        self.__add_connection_by_index(mechanical_connections, "output", append_to_support_block=0)

        self.__add_connection_by_index(electrical_connections, "input")
        self.__add_connection_by_index(electrical_connections, "output")

    def __add_connection_by_index(self, input_list: ETree.Element, connection_name, append_to_support_block = None):

        if connection_name == "input":

            is_input = True

        else:

            is_input = False

        for connection in input_list.findall(connection_name):

            new_conn = self.main_class.find_connection_by_index(float(connection.get("index")))

            if new_conn is not None:

                self.add_connection(new_conn, is_input, append_to_support_block=append_to_support_block)

    @classmethod
    def return_EES_needed_index(cls):

        # Alternator has multiple input and outputs for exergy flux:

        return_dict = {"global output": [0, False],
                       "mechanical input": [1, True],
                       "mechanical output": [2, True],
                       "electrical input": [3, True],
                       "electrical output": [4, True]}

        return return_dict

    @classmethod
    def return_EES_base_equations(cls):

        # WARNING: This methods must be overloaded in subclasses!!
        # This methods returns a dictionary that contain a list of streams that have to be present in the EES text
        # definition.

        return dict()

    @property
    def exergy_balance(self):

        exergy_balance = 0

        for conn in self.input_connections:

            if conn == self.support_block[0].connection_with_main:
                exergy_balance += conn.exergy_value * self.efficiency

            else:
                exergy_balance += conn.exergy_value

        for conn in self.output_connections:
            exergy_balance -= conn.exergy_value

        return exergy_balance

    @property
    def can_be_removed_in_pf_definition(self):
        return False

    def return_other_zone_connections(self, zone_type, input_connection):

        # Alternator is connected only to energy streams, hence it is not interested in the zones generation process
        return list()


import numpy
from copy import deepcopy


class MatrixAnalyzer:

    def __init__(self, matrix: numpy.ndarray, vector: numpy.ndarray, main_matrix=None):

        self.matrix = deepcopy(matrix)
        self.vector = vector

        self.main_matrix = main_matrix
        self.sub_matrix = None

        self.has_assignment = False

        self.n_row = self.matrix.shape[0]
        self.n_col = self.matrix.shape[1]

        self.elements = list()

        self.vector_list = None
        self.columns = MatrixLineList(contains_rows=False)
        self.rows = MatrixLineList(contains_rows=True)

        self.__initialize_matrix()

        self.rows.check()
        self.columns.check()
        self.__check_assignments()

        self.__ill_conditioned_matrix_warning = False

    def __initialize_matrix(self):

        for i in range(self.n_row):

            row_list = list()
            for j in range(self.n_col):
                new_element = MatrixElement(self.matrix[i, j], [i, j])
                row_list.append(new_element)

            self.elements.append(row_list)
            self.rows.add_to_list(row_list, i)

        for j in range(self.n_col):

            col_list = list()
            for i in range(self.n_row):
                element = self.elements[i][j]
                col_list.append(element)

            self.columns.add_to_list(col_list, j)

        __tmp_vector_list = list()

        for i in range(self.n_row):
            __tmp_vector_list.append(MatrixElement(self.vector[i], [i, -1]))

        self.vector_list = MatrixColumn(__tmp_vector_list, -1)

    def __check_assignments(self):

        for row in self.rows:

            if row.is_assignment:
                self.has_assignment = True

    def solve(self):

        if self.has_assignment:

            non_assignment_rows_indices = self.__calculate_assignment()

            if len(non_assignment_rows_indices) > 0:

                __return = self.__identify_new_matrix_and_vector(non_assignment_rows_indices)
                non_solved_columns = self.columns.non_solved

                self.sub_matrix = MatrixAnalyzer(__return["new_matrix"], __return["new_vector"], self)
                sub_solution = self.sub_matrix.solve()

                for i in range(len(non_solved_columns)):
                    non_solved_columns[i].set_solution(sub_solution[i])

        else:

            try:

                sol = numpy.linalg.solve(self.matrix, self.vector)

            except:

                sol = numpy.linalg.lstsq(self.matrix, self.vector)
                sol = sol[0]

                self.__ill_conditioned_matrix_warning = True

            for i in range(len(self.columns)):
                self.columns[i].set_solution(sol[i])

        return self.solution

    def __calculate_assignment(self) -> list:

        non_assignment_rows_indices = list()

        for row in self.rows:

            if row.is_assignment:

                assignment_elem = row.non_zero_element[0]
                assignment_vector = self.vector_list.get_from_list(assignment_elem.initial_position[0])
                solution = assignment_vector.value / assignment_elem.value

                assignment_column = self.columns.get_from_list(assignment_elem.initial_position[1])
                assignment_column.set_solution(solution)

            else:

                non_assignment_rows_indices.append(row.initial_position)

        return non_assignment_rows_indices

    def __identify_new_matrix_and_vector(self, non_assignment_rows_indices):

        # prepare new sub_matrix
        n_elements = len(non_assignment_rows_indices)
        new_matrix = numpy.zeros((n_elements, n_elements))
        new_vector = self.vector_list.return_numpy_array(indices=non_assignment_rows_indices)

        i = 0
        for column in self.columns:

            row_element = column.return_numpy_array(indices=non_assignment_rows_indices)

            if column.is_solved:

                new_vector -= row_element * column.solution

            else:

                new_matrix[:, i] = row_element
                i += 1

        return {"new_matrix": new_matrix,
                "new_vector": new_vector}

    @property
    def is_solvable(self):

        for element in [self.columns.empty_lines, self.columns.dependent_lines,
                        self.rows.empty_lines, self.rows.dependent_lines]:

            if not len(element) == 0:
                return False

        return self.n_row == self.n_col

    @property
    def solution(self):

        sol_vector = numpy.zeros(len(self.vector))

        for i in range(len(self.columns)):

            solution = self.columns[i].solution

            if solution is not None:
                sol_vector[i] = solution

        return sol_vector

    @property
    def is_ill_conditioned(self):

        if self.sub_matrix is None:

            return self.__ill_conditioned_matrix_warning

        else:

            return self.sub_matrix.is_ill_conditioned


class MatrixElement:

    def __init__(self, value: float, initial_position: list):
        self.value = value
        self.solution = None
        self.initial_position = initial_position

        self.row = None
        self.col = None

    @property
    def is_empty(self) -> bool:
        return self.value == 0


class MatrixLine:

    def __init__(self, vector: list[MatrixElement], initial_position):

        self.vector = vector
        self.module = self.__vector_module()
        self.initial_position = initial_position

        self.is_row = None
        self.solution = None

        self.non_zero = self.__count_non_zero_elements()

    def __vector_module(self) -> float:

        module = 0

        for element in self.vector:
            module += pow(element.value, 2)

        return pow(module, 0.5)

    def __count_non_zero_elements(self) -> int:

        counter = 0

        for element in self.vector:

            if not element.is_empty:
                counter += 1

        return counter

    def return_numpy_array(self, indices=None) -> numpy.ndarray:

        if indices is None:

            return_vector = numpy.zeros(len(self.vector))

            for i in range(len(self.vector)):
                return_vector[i] = self.vector[i].value

        else:

            i = 0
            return_vector = numpy.zeros(len(indices))

            for index in indices:
                return_vector[i] = self.vector[index].value
                i += 1

        return return_vector

    def get_from_list(self, initial_position):

        if self.is_row:
            i = 1
        else:
            i = 0

        for element in self.vector:

            if element.initial_position[i] == initial_position:
                return element

        return None

    @property
    def non_zero_element(self):

        return_list = list()

        for element in self.vector:

            if not element.is_empty:
                return_list.append(element)

        return return_list

    @property
    def is_empty(self):

        return self.module == 0

    @property
    def dim(self):

        return len(self.vector)

    def __gt__(self, other):

        # enables comparison
        # self > other

        return self.module < other.module

    def __lt__(self, other):

        # enables comparison
        # self < other

        return self.module > other.module

    def __le__(self, other):

        return not self.__gt__(other)

    def __ge__(self, other):

        return not self.__lt__(other)

    def __eq__(self, other):

        if self.dim == other.dim and self.is_row == other.is_row:

            if self.non_zero == other.non_zero:

                for i in range(self.dim):

                    if self.vector[i].value / self.module != other.vector[i].value / other.module:
                        return False

                return True

        return False


class MatrixColumn(MatrixLine):

    def __init__(self, vector: list[MatrixElement], initial_position):

        super().__init__(vector, initial_position)

        self.is_row = False

        for elem in vector:
            elem.col = self

    def set_solution(self, solution):

        self.solution = solution
        for elem in self.vector:
            elem.solution = solution

    @property
    def is_solved(self):

        return self.solution is not None


class MatrixRow(MatrixLine):

    def __init__(self, vector: list[MatrixElement], initial_position):
        super().__init__(vector, initial_position)
        self.is_row = True

        for elem in vector:
            elem.row = self

    @property
    def is_assignment(self):
        return self.non_zero == 1


class MatrixLineList(list):

    def __init__(self, contains_rows):

        super().__init__()
        self.contains_rows = contains_rows

        self.empty_lines = list()
        self.dependent_lines = list()

    def get_from_list(self, initial_position):

        for element in self:

            if element.initial_position == initial_position:
                return element

        return None

    def add_to_list(self, vector: list[MatrixElement], initial_position: int):

        new_element = self.get_from_list(initial_position)

        if new_element is None:

            if self.contains_rows:
                new_element = MatrixRow(vector, initial_position)

            else:
                new_element = MatrixColumn(vector, initial_position)

            self.append(new_element)

        return new_element

    def check(self):

        self.empty_lines = list()
        self.dependent_lines = list()

        for i in range(len(self)):

            if self[i].is_empty:

                self.empty_lines.append({i: self[i]})

            else:

                for j in range(i + 1, len(self)):

                    if self[i] == self[j]:
                        self.dependent_lines.append({str(i) + "; " + str(j): [self[i], self[j]]})

    @property
    def non_solved(self):

        return_list = list()

        for element in self:

            if not element.is_solved:
                return_list.append(element)

        return return_list

    @property
    def sorted_indices(self) -> list:

        sorted_indices = list()
        self.sort()

        for element in self:

            sorted_indices.append(element.initial_position)

        return sorted_indices


from EEETools import costants


class Zones:

    def __init__(self, array_handler):

        self.array_handler = array_handler
        self.__initialize_zones()

    def __initialize_zones(self):

        self.zones = dict()

        for zone_type in costants.ZONES:

            zone_list = list()
            first_connection =  self.__find_new_first_connection(zone_list)

            while first_connection is not None:
                zone_list.append(Zone(zone_type, first_connection))
                first_connection = self.__find_new_first_connection(zone_list)

            self.zones.update({zone_type: zone_list})

    def __find_new_first_connection(self, zone_list):

        for connection in self.array_handler.connection_list:

            if connection.is_fluid_stream:

                found = False

                for zone in zone_list:

                    if zone.contains(connection):
                        found = True
                        break

                if not found:
                    return connection

        return None


class Zone:

    def __init__(self, type, first_connection):

        self.connections = list()
        self.type = type
        self.is_defined = False

        self.add_connections(first_connection)

    def add_connections(self, first_connection):

        self.add_connection(first_connection)

    def add_connection(self, connection):

        if connection not in self.connections:

            if not connection.zones[self.type] is None:
                connection.zones[self.type].remove_connection(connection)

            connection.add_zone(self)
            self.connections.append(connection)

            for new_connection in connection.return_other_zone_connections(self):
                self.add_connection(new_connection)

    def remove_connection(self, connection):
        self.connections.remove(connection)
        connection.zones[self.type] = None

    def contains(self, connection):

        return connection in self.connections



from EEETools.Tools.CostCorrelations.cost_correlation_classes import CostCorrelationHandler
from EEETools.Tools.CostCorrelations.cost_correlation_classes import AbstractCostCorrelation
from EEETools.Tools.CostCorrelations.cost_correlation_classes import Parameter
from EEETools.Tools.GUIElements.gui_base_classes import *
from PyQt5.QtGui import QKeySequence, QFont, QDoubleValidator
from shutil import copy2 as copy_file
from PyQt5.QtWidgets import *
from PyQt5.QtCore import Qt
import os


class CostCorrelationFilesManager(AbstractFilesManager):

    def __init__(self, flags=None, *args, **kwargs):

        super().__init__(flags, *args, **kwargs)

        self.title = "Cost Correlation Manager"
        self.modules_handler = CostCorrelationHandler()

    def on_button_add_pressed(self):

        if not os.path.isdir(self.selected_element_path):

            self.dir_name = os.path.dirname(self.selected_element_path)

        else:

            self.dir_name = self.selected_element_path

        warning_text = "Select a correlation format"

        self.add_new_dialog = WarningDialog(self)
        self.add_new_dialog.set_label_text(warning_text)
        self.add_new_dialog.set_buttons_on_click(self.on_add_new_dialog_clicked)

        self.add_combo_box = QComboBox()
        self.add_combo_box.addItems(self.modules_handler.list_modules())

        combo_box_layout = QHBoxLayout()
        combo_box_layout.addWidget(self.add_combo_box)
        self.add_new_dialog.set_layout(combo_box_layout)

        self.add_new_dialog.set_button_text("Add", set_button_yes=True)
        self.add_new_dialog.set_button_text("Cancel", set_button_yes=False)

        self.add_new_dialog.show()

    def on_add_new_dialog_clicked(self):

        sender = self.sender()

        if sender.is_yes_button:

            __sub_class_name = self.add_combo_box.currentText()
            __sub_class = self.modules_handler.import_correct_sub_class(__sub_class_name)()

            component_type = self.__get_component_type()
            __sub_class.component_type = component_type

            modify_widget = __sub_class.edit_correlation_widget
            modify_widget.show()

        self.add_new_dialog.close()

    def on_button_modify_pressed(self):

        __sub_class = self.modules_handler.open_file(self.selected_element_path)

        modify_widget = __sub_class.edit_correlation_widget
        modify_widget.show()

    def on_button_delete_pressed(self):

        dir_name = os.path.dirname(self.selected_element_path)

        dir_name = os.path.basename(dir_name)
        filename = os.path.basename(self.selected_element_path)

        warning_text = "Do you really want to delete file "
        warning_text += os.path.join(dir_name, filename)
        warning_text += "?"

        self.delete_warning_dialog = WarningDialog(self)
        self.delete_warning_dialog.set_label_text(warning_text)
        self.delete_warning_dialog.set_buttons_on_click(self.on_delete_warning_clicked)
        self.delete_warning_dialog.set_button_text("Yes", set_button_yes=True)
        self.delete_warning_dialog.set_button_text("No", set_button_yes=False)

        self.delete_warning_dialog.show()

    def on_delete_warning_clicked(self):

        sender = self.sender()

        if sender.is_yes_button:
            if os.path.exists(self.selected_element_path):
                os.remove(self.selected_element_path)

        self.__enable_buttons()
        self.delete_warning_dialog.close()

    def on_button_rename_pressed(self):

        dir_name = os.path.dirname(self.selected_element_path)

        dir_name = os.path.basename(dir_name)
        filename = os.path.basename(self.selected_element_path)

        label_text = "insert new name for file "
        label_text += os.path.join(dir_name, filename)

        self.rename_line_edit = QLineEdit()
        hint_label = QLabel("new name:")

        hLayout = QHBoxLayout()
        hLayout.addWidget(hint_label)
        hLayout.addWidget(self.rename_line_edit)

        self.rename_warning_dialog = WarningDialog(self)
        self.rename_warning_dialog.set_layout(hLayout)
        self.rename_warning_dialog.set_label_text(label_text)
        self.rename_warning_dialog.set_buttons_on_click(self.on_rename_warning_clicked)
        self.rename_warning_dialog.set_button_text("Rename", set_button_yes=True)
        self.rename_warning_dialog.set_button_text("Cancel", set_button_yes=False)

        self.rename_warning_dialog.show()

    def on_rename_warning_clicked(self):

        sender = self.sender()

        if sender.is_yes_button:
            new_name = self.rename_line_edit.text()
            self.__rename_file(new_name, self.selected_element_path)
            self.__enable_buttons()

        self.rename_warning_dialog.close()

    def on_button_duplicate_pressed(self):

        dir_name = os.path.dirname(self.selected_element_path)

        dir_name = os.path.basename(dir_name)
        filename = os.path.basename(self.selected_element_path)

        warning_text = "Do you really want to duplicate file "
        warning_text += os.path.join(dir_name, filename)
        warning_text += "?"

        self.duplicate_warning_dialog = WarningDialog(self)

        self.duplicate_warning_dialog.set_label_text(warning_text)

        self.duplicate_warning_dialog.set_button_text("Yes", set_button_yes=True)
        self.duplicate_warning_dialog.set_button_text("No", set_button_yes=False)
        self.duplicate_warning_dialog.set_buttons_on_click(self.on_duplicate_warning_pressed)

        self.duplicate_warning_dialog.show()

    def on_duplicate_warning_pressed(self):

        sender = self.sender()

        if sender.is_yes_button:

            if os.path.exists(self.selected_element_path):

                dir_name = os.path.dirname(self.selected_element_path)
                filename = os.path.basename(self.selected_element_path)

                if ".dat" in filename:
                    filename = filename.split(".dat")[0]
                    file_extension = ".dat"

                elif ".xml" in filename:
                    filename = filename.split(".xml")[0]
                    file_extension = ".xml"

                else:

                    file_extension = ""

                i = 1
                new_path = os.path.join(dir_name, filename + "(1)" + file_extension)

                while os.path.exists(new_path):
                    i += 1
                    new_path = os.path.join(dir_name, filename + "(" + str(i) + ")" + file_extension)

                copy_file(self.selected_element_path, new_path)

        self.duplicate_warning_dialog.close()

    def __rename_file(self, new_name, file_path):

        self.modules_handler.rename_file(new_name, file_path)
        super(CostCorrelationFilesManager, self).__rename_file(new_name, file_path)

    def __get_component_type(self):

        if not os.path.isdir(self.selected_element_path):

            dir_name = os.path.dirname(self.selected_element_path)

        else:

            dir_name = self.selected_element_path

        return os.path.basename(dir_name)


class BaseEditCorrelationWidget(QDialog):

    def __init__(self, cost_correlation_class: AbstractCostCorrelation):

        super().__init__()

        self.title = ""
        self.cost_correlation = cost_correlation_class

        self.tab_widget = BaseEditCorrelationTabWidget()
        self.buttons_layout = QHBoxLayout()

        self.button_save = QPushButton()
        self.button_new = QPushButton()

        self.layout_is_set = False

    def show(self):

        if not self.layout_is_set:
            self.__init_main_layout()

        super(BaseEditCorrelationWidget, self).show()

    def enable_buttons(self):

        self.button_save.setEnabled(self.is_ready_for_saving)

    def __init_main_layout(self):

        self.setWindowTitle(self.title)
        self.__init_button_layout()

        __button_widget = QWidget()
        __button_widget.setLayout(self.buttons_layout)

        main_layout = QVBoxLayout()

        main_layout.addWidget(self.tab_widget)
        main_layout.addWidget(__button_widget)

        self.setLayout(main_layout)
        self.layout_is_set = True

    def __init_tab_layout(self):

        self.coefficient_widget = CoefficientWidget(self.cost_correlation.coefficient_dict, self)

    def __init_button_layout(self):

        self.button_save.setText("Save (CTRL-S)")
        self.button_save.clicked.connect(self.on_button_pressed)

        self.button_new.setText("New (CTRL-N)")
        self.button_new.clicked.connect(self.on_button_pressed)

        self.__set_buttons_shortcut()

        self.buttons_layout = QHBoxLayout()
        self.buttons_layout.addWidget(self.button_save)
        self.buttons_layout.addWidget(self.button_new)

        self.enable_buttons()

    def __set_buttons_shortcut(self):

        sequence_save = QKeySequence(Qt.CTRL + Qt.Key_S)
        self.button_save.setShortcut(sequence_save)

        sequence_new = QKeySequence(Qt.CTRL + Qt.Key_N)
        self.button_new.setShortcut(sequence_new)

    @property
    def is_ready_for_saving(self):
        """method to be overloaded"""
        return False

    # <------------------------------------------------->
    # <------------ BUTTON ON CLICK METHODS ------------>
    # <------------------------------------------------->

    def on_button_pressed(self):

        sender = self.sender()

        if "Save" in sender.text():

            warning_text = "Do you want to save the file? Data will be overwritten"
            self.on_click_function = self.__save

        else:

            warning_text = "By clicking \"Continue\" all your progress will be lost"
            self.on_click_function = self.__reset_dialog

        self.warning_dialog = WarningDialog(self)

        self.warning_dialog.set_label_text(warning_text)
        self.warning_dialog.set_buttons_on_click(self.on_warning_clicked)

        self.warning_dialog.set_button_text("Continue", set_button_yes=True)
        self.warning_dialog.set_button_text("Cancel", set_button_yes=False)

        self.warning_dialog.show()

    def on_warning_clicked(self):

        sender = self.sender()
        if sender.is_yes_button:
            self.on_click_function()

        self.warning_dialog.close()

    def __save(self):

        self.__update()
        self.cost_correlation.export_to_file()

    def __reset_dialog(self):
        raise NotImplementedError

    def __update(self):
        raise NotImplementedError

    # <------------------------------------------------->
    # <------------ WIDGETS GETTER AND SETTER ----------->
    # <------------------------------------------------->

    @property
    def header_widget(self):
        return self.tab_widget.tab_main

    @header_widget.setter
    def header_widget(self, input_widget):
        self.tab_widget.add_widget(input_widget, "Main")

    @property
    def parameter_widget(self):
        return self.tab_widget.tab_parameters

    @parameter_widget.setter
    def parameter_widget(self, input_widget):
        self.tab_widget.add_widget(input_widget, "Parameters")

    @property
    def coefficient_widget(self):
        return self.tab_widget.tab_coefficient

    @coefficient_widget.setter
    def coefficient_widget(self, input_widget):
        self.tab_widget.add_widget(input_widget, "Coefficients")


class BaseEditCorrelationTabWidget(QTabWidget):

    def __init__(self):

        super().__init__()

        self.tab_main = None
        self.tab_parameters = None
        self.tab_coefficient = None

    def add_widget(self, widget: QWidget, widget_type):

        found = True

        if widget_type == "Main" and self.tab_main is None:
            self.tab_main = widget

        elif  widget_type == "Parameters" and self.tab_parameters is None:
            self.tab_parameters = widget

        elif  widget_type == "Coefficients" and self.tab_coefficient is None:
            self.tab_coefficient = widget

        else:
            found = False

        if found:
            self.addTab(widget, widget_type)


class ParametersWidget(QWidget):

    def __init__(self, parameter_dict: dict, parent_class: BaseEditCorrelationWidget):

        super().__init__()
        self.splitter = QSplitter()
        self.parameter_dict = parameter_dict

        self.list_view = None
        self.parameters_widgets = dict()

    def init_layout(self):

        self.list_view = QListView()
        self.list_view.clicked.connect(self.__on_list_view_clicked)

        row = 0
        for key in self.parameter_dict.keys():

            self.list_view.insertItem(row, key)
            row += 1


    #
    # def __on_list_view_clicked(self):
    #
    #
    #
    # def __show_currect_widget(self, item_key: str):
    #
    #


    @property
    def is_ready_for_saving(self):

        for edit_text in [self.unit_edit, self.note_edit, self.range_low_edit, self.range_high_edit]:

            if edit_text.text() == "":
                return False

        return True


class SingleParameterWidget(QWidget):

    def __init__(self, parameter_class: Parameter, parent_class: BaseEditCorrelationWidget):

        super().__init__()

        self.set_edit_layout = False
        self.parameter = parameter_class
        self.parent_class = parent_class

        self.main_layout = QVBoxLayout()
        self.grid_layout = QGridLayout()

    def init_layout(self):

        if self.set_edit_layout:

            self.__init_edit_objects()
            self.__init_edit_layout()

        else:
            self.__init_set_objects()
            self.__init_set_layout()

        self.setLayout(self.main_layout)

    def __init_edit_objects(self):

        self.unit_edit = QLineEdit()
        self.note_edit = QTextEdit()
        self.range_low_edit = QLineEdit()
        self.range_high_edit = QLineEdit()

        self.range_low_edit.setValidator(QDoubleValidator())
        self.range_high_edit.setValidator(QDoubleValidator())

        self.unit_edit.setText(self.parameter.measure_unit)
        self.note_edit.setText(self.parameter.note)
        self.range_low_edit.setText(str(self.parameter.range.low))
        self.range_high_edit.setText(str(self.parameter.range.high))

        self.unit_edit.setObjectName("unit_edit")
        self.note_edit.setObjectName("note_edit")
        self.range_low_edit.setObjectName("range_low_edit")
        self.range_high_edit.setObjectName("range_high_edit")

        self.unit_edit.textChanged.connect(self.on_text_edit_change)
        self.note_edit.textChanged.connect(self.on_text_edit_change)
        self.range_low_edit.textChanged.connect(self.on_text_edit_change)
        self.range_high_edit.textChanged.connect(self.on_text_edit_change)

    def __init_set_objects(self):

        self.value_edit = QLineEdit()
        self.unit_edit = QLabel()
        self.note_edit = QLabel()
        self.range_low_edit = QLabel()
        self.range_high_edit = QLabel()

        self.note_edit.setWordWrap(True)
        self.value_edit.setObjectName("value_edit")
        self.value_edit.textChanged.connect(self.on_text_edit_change)
        self.value_edit.setValidator(QDoubleValidator())

        self.value_edit.setText(str(self.parameter.value))
        self.unit_edit.setText(self.parameter.measure_unit)
        self.note_edit.setText(self.parameter.note)
        self.range_low_edit.setText(str(self.parameter.range.low))
        self.range_high_edit.setText(str(self.parameter.range.high))

    def __init_edit_layout(self):

        # <----------- FIRST ROW ----------->

        title_label_input = QLabel()
        title_label_input.setText(self.parameter.name)
        title_label_input.setFont(QFont('Helvetica 30 bold'))

        self.grid_layout.addWidget(title_label_input, 0, 0)

        # <----------- SECOND ROW ----------->

        measure_unit_label = QLabel()
        measure_unit_label.setText("Measurement Unit")

        self.grid_layout.addWidget(measure_unit_label, 1, 0)
        self.grid_layout.addWidget(self.unit_edit, 1, 1)

        # <----------- THIRD ROW ----------->

        range_label = QLabel()
        range_label.setText("Range")

        range_layout = QHBoxLayout()
        range_layout.addWidget(self.range_low_edit)
        range_layout.addWidget(QLabel("->"))
        range_layout.addWidget(self.range_high_edit)

        range_widget = QWidget()
        range_widget.setLayout(range_layout)

        self.grid_layout.addWidget(range_label, 2, 0)
        self.grid_layout.addWidget(range_widget, 2, 1)

        # <----------- FINAL ROW ----------->

        grid_layout_widget = QWidget()
        grid_layout_widget.setLayout(self.grid_layout)

        self.main_layout.addWidget(grid_layout_widget)
        self.main_layout.addWidget(self.note_edit)

    def __init_set_layout(self):

        # <----------- FIRST ROW ----------->

        title_label_input = QLabel()
        title_label_input.setText(self.parameter.name)
        title_label_input.setFont(QFont('Helvetica 30 bold'))

        self.grid_layout.addWidget(title_label_input, 0, 0)

        # <----------- SECOND ROW ----------->

        measure_unit_label = QLabel()
        measure_unit_label.setText("Measurement Unit")

        self.grid_layout.addWidget(measure_unit_label, 1, 0)
        self.grid_layout.addWidget(self.unit_edit, 1, 1)

        # <----------- THIRD ROW ----------->

        range_label = QLabel()
        range_label.setText("Range")

        range_layout = QHBoxLayout()
        range_layout.addWidget(self.range_low_edit)
        range_layout.addWidget(QLabel("->"))
        range_layout.addWidget(self.range_high_edit)

        range_widget = QWidget()
        range_widget.setLayout(range_layout)

        self.grid_layout.addWidget(range_label, 2, 0)
        self.grid_layout.addWidget(range_widget, 2, 1)

        # <----------- FINAL ROW ----------->

        grid_layout_widget = QWidget()
        grid_layout_widget.setLayout(self.grid_layout)

        self.main_layout.addWidget(grid_layout_widget)
        self.main_layout.addWidget(self.note_edit)

    def on_text_edit_change(self):

        name = self.sender().objectName()

        if name == "value_edit":

            try:
                self.parameter.value = float(self.value_edit.text())
            except:
                self.parameter.value = 0.0

        elif name == "unit_edit":

            self.parameter.measure_unit = self.unit_edit.text()

        elif name == "note_edit":

            self.parameter.note = self.note_edit.toPlainText()

        elif name == "range_low_edit" or name == "range_high_edit":

            try:
                low = float(self.range_low_edit.text())
            except:
                low = 0.0

            try:
                high = float(self.range_high_edit.text())
            except:
                high = 0.0

            self.parameter.range.initialize([low, high])

        self.parent_class.enable_buttons()

    @property
    def is_ready_for_saving(self):

        for edit_text in [self.unit_edit, self.note_edit, self.range_low_edit, self.range_high_edit]:

            if edit_text.text() == "":
                return False

        return True


class CoefficientWidget(QWidget):

    def __init__(self, coefficient_dict: dict, parent_class: BaseEditCorrelationWidget):

        super().__init__()

        self.coefficient_dict = coefficient_dict
        self.parent_class = parent_class

        self.main_layout = QGridLayout()
        self.widget_dict = dict()

        self.__init_layout()

    def __init_layout(self):

        row = 0

        for key in self.coefficient_dict.keys():

            self.__extend_grid_layout(key, row)
            row += 1

        self.setLayout(self.main_layout)

    def __extend_grid_layout(self, key, row):

        col = 1
        coefficient = self.coefficient_dict[key]

        title_label_input = QLabel()
        title_label_input.setText("<b>"+ key + ":<b>")
        title_label_input.setAlignment(Qt.AlignRight)

        self.grid_layout.addWidget(title_label_input, row, col)

        i = 0
        col += 1
        first_col = col
        edit_test_list = list()

        while i < coefficient["array size"]:

            col = first_col + i

            edit_text = QLineEdit()
            edit_text.setValidator(QDoubleValidator())
            edit_text.textChanged.connect(self.on_text_edit_change)

            edit_test_list.append(edit_text)
            self.grid_layout.addWidget(edit_text, row, col)

            i += 1

        self.widget_dict.update({key: edit_test_list})

    def on_text_edit_change(self):

        self.parent_class.enable_buttons()

    @property
    def coefficient_values(self):

        return_dict = dict()

        for key in self.widget_dict.keys():

            return_list = list()

            for edit_text in self.widget_dict[key]:

                try:
                    return_list.append(float(edit_text.text()))

                except:
                    return_list.append(0.0)

            return_dict.update({key: return_list})

        return return_dict

    @property
    def is_ready_for_saving(self):

        for key in self.widget_dict.keys():

            for edit_text in self.widget_dict[key]:

                if edit_text.text() == "":

                    return False

        return True